/*
**  vmware-console.c -- VMWare GSX Server Remote Console binary RTLD wrapper
*/

#define CONFIG_OLD_PATHSTR "/etc/vmware-console/config"
#define CONFIG_NEW_ENVVAR  "VMWARE_CONSOLE_CONFIG"

#define fopen remove_fopen
#include <stdio.h>
#undef fopen
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#if defined(PATH_MAX)
#define PATH_ADJUST_MAXLEN PATH_MAX
#elif defined(MAXPATHLEN)
#define PATH_ADJUST_MAXLEN MAXPATHLEN
#else
#define PATH_ADJUST_MAXLEN 1024
#endif

static void path_adjust(const char *func, char *buf, size_t buflen, const char *path)
{
    size_t pathlen;
    char *cp;

    /* special case first */
    if (path == NULL) {
        buf[0] = '\0';
        return;
    }

    /* take over path into buffer */
    pathlen = strlen(path);
    if (pathlen > buflen)
        pathlen = buflen;
    strncpy(buf, path, pathlen);
    buf[pathlen] = '\0';

    /* apply path adjustments */
    if (strcmp(buf, CONFIG_OLD_PATHSTR) == 0) {
        if ((cp = getenv(CONFIG_NEW_ENVVAR)) != NULL)
            strcpy(buf, cp);
    }

    return;
}

#include <features.h>
#define __USE_GNU
#include <dlfcn.h>
#undef __USE_GNU
#include <fcntl.h>

static int
myopen(
    const char *func_name,
    int (**func_ptr)(const char *, int, ...),
    char *buf,
    size_t buflen,
    const char *path,
    int flags,
    va_list ap)
{
    int rv;
    int mode;

    /* initially resolve original open(2) like function */
    if (*func_ptr == NULL) {
        if ((*func_ptr = dlsym(RTLD_NEXT, func_name)) == NULL) {
            fprintf(stderr, "vmware-console.so: unable to resolve function \"%s\"\n", func_name);
            abort();
        }
    }

    /* adjust path */
    path_adjust(func_name, buf, buflen, path);

    /* execute original function */
    if (flags & O_CREAT) {
        mode = (int)va_arg(ap, int);
        rv = (*func_ptr)(buf, flags, mode);
    }
    else
        rv = (*func_ptr)(buf, flags);

    /* pass-through return value */
    return rv;
}

#define genstub_open(name) \
int name(const char *path, int flags, ...) \
{ \
    static int (*func)(const char *, int, ...) = NULL; \
    char path_adjusted[PATH_ADJUST_MAXLEN]; \
    va_list ap; \
    int rv; \
\
    va_start(ap, flags); \
    rv = myopen(#name, &func, path_adjusted, sizeof(path_adjusted), path, flags, ap); \
    va_end(ap); \
    return rv; \
}

genstub_open(open)
genstub_open(open64)
genstub_open(__open)
genstub_open(__open64)
genstub_open(__libc_open)
genstub_open(__libc_open64)

static int
mystat(
    const char *func_name,
    int (**func_ptr)(const char *, struct stat *),
    char *buf,
    size_t buflen,
    const char *path,
    struct stat *sb)
{
    int rv;

    /* initially resolve original stat(2) like function */
    if (*func_ptr == NULL) {
        if ((*func_ptr = dlsym(RTLD_NEXT, func_name)) == NULL) {
            fprintf(stderr, "vmware-console.so: unable to resolve function \"%s\"\n", func_name);
            abort();
        }
    }

    /* adjust path */
    path_adjust(func_name, buf, buflen, path);

    /* execute original function */
    rv = (*func_ptr)(buf, sb);

    /* pass-through return value */
    return rv;
}

#define genstub_stat(name) \
int name(const char *path, struct stat *sb) \
{ \
    static int (*func)(const char *, struct stat *sb) = NULL; \
    char path_adjusted[PATH_ADJUST_MAXLEN]; \
    int rv; \
\
    rv = mystat(#name, &func, path_adjusted, sizeof(path_adjusted), path, sb); \
    return rv; \
}

genstub_stat(stat)
genstub_stat(stat64)
genstub_stat(__stat)
genstub_stat(__stat64)
genstub_stat(__libc_stat)
genstub_stat(__libc_stat64)

static int
myaccess(
    const char *func_name,
    int (**func_ptr)(const char *, int mode),
    char *buf,
    size_t buflen,
    const char *path,
    int mode)
{
    int rv;

    /* initially resolve original access(2) like function */
    if (*func_ptr == NULL) {
        if ((*func_ptr = dlsym(RTLD_NEXT, func_name)) == NULL) {
            fprintf(stderr, "vmware-console.so: unable to resolve function \"%s\"\n", func_name);
            abort();
        }
    }

    /* adjust path */
    path_adjust(func_name, buf, buflen, path);

    /* execute original function */
    rv = (*func_ptr)(buf, mode);

    /* pass-through return value */
    return rv;
}

#define genstub_access(name) \
int name(const char *path, int mode) \
{ \
    static int (*func)(const char *, int mode) = NULL; \
    char path_adjusted[PATH_ADJUST_MAXLEN]; \
    int rv; \
\
    rv = myaccess(#name, &func, path_adjusted, sizeof(path_adjusted), path, mode); \
    return rv; \
}

genstub_access(access)
genstub_access(access64)
genstub_access(__access)
genstub_access(__access64)
genstub_access(__libc_access)
genstub_access(__libc_access64)

static FILE *
myfopen(
    const char *func_name,
    FILE *(**func_ptr)(const char *, const char *),
    char *buf,
    size_t buflen,
    const char *path,
    const char *mode)
{
    FILE *rv;

    /* initially resolve original fopen(3) like function */
    if (*func_ptr == NULL) {
        if ((*func_ptr = dlsym(RTLD_NEXT, func_name)) == NULL) {
            fprintf(stderr, "vmware-console.so: unable to resolve function \"%s\"\n", func_name);
            abort();
        }
    }

    /* adjust path */
    path_adjust(func_name, buf, buflen, path);

    /* execute original function */
    rv = (*func_ptr)(buf, mode);

    /* pass-through return value */
    return rv;
}

#define genstub_fopen(name) \
FILE *name(const char *path, const char *mode) \
{ \
    static FILE *(*func)(const char *, const char *mode) = NULL; \
    char path_adjusted[PATH_ADJUST_MAXLEN]; \
    FILE *rv; \
\
    rv = myfopen(#name, &func, path_adjusted, sizeof(path_adjusted), path, mode); \
    return rv; \
}

genstub_fopen(fopen)
genstub_fopen(__fopen)
genstub_fopen(__libc_fopen)

static int
myxstat(
    const char *func_name,
    int (**func_ptr)(int, const char *, struct stat *),
    char *buf,
    size_t buflen,
    int ver, const char *path, struct stat *sb)
{
    int rv;

    /* initially resolve original __xstat(2) like function */
    if (*func_ptr == NULL) {
        if ((*func_ptr = dlsym(RTLD_NEXT, func_name)) == NULL) {
            fprintf(stderr, "vmware-console.so: unable to resolve function \"%s\"\n", func_name);
            abort();
        }
    }

    /* adjust path */
    path_adjust(func_name, buf, buflen, path);

    /* execute original function */
    rv = (*func_ptr)(ver, buf, sb);

    /* pass-through return value */
    return rv;
}

#define genstub_xstat(name) \
int name(int ver, const char *path, struct stat *sb) \
{ \
    static int (*func)(int, const char *, struct stat *sb) = NULL; \
    char path_adjusted[PATH_ADJUST_MAXLEN]; \
    int rv; \
\
    rv = myxstat(#name, &func, path_adjusted, sizeof(path_adjusted), ver, path, sb); \
    return rv; \
}

genstub_xstat(xstat)
genstub_xstat(xstat64)
genstub_xstat(__xstat)
genstub_xstat(__xstat64)
genstub_xstat(__libc_xstat)
genstub_xstat(__libc_xstat64)

